# The opposite of casting. See lightbus.utilities.casting
from collections import OrderedDict
from datetime import datetime, date
from decimal import Decimal
from enum import Enum
from uuid import UUID
from base64 import b64encode

from lightbus.exceptions import DeformError
from lightbus.utilities.frozendict import frozendict
from lightbus.utilities.type_checks import is_namedtuple, is_dataclass, isinstance_safe


def deform_to_bus(value):
    """Convert value into one which can be safely serialised
    and encoded onto the bus

    The opposite of cast_to_signature()
    """
    # pylint: disable=too-many-return-statements,unidiomatic-typecheck,import-outside-toplevel
    if value is None:
        return value
    elif hasattr(value, "__to_bus__"):
        return deform_to_bus(value.__to_bus__())
    elif isinstance(value, OrderedDict):
        return deform_to_bus(dict(value))
    elif isinstance_safe(value, dict):
        new_dict = {}
        for dict_key, dict_value in value.items():
            new_dict[dict_key] = deform_to_bus(dict_value)
        return new_dict
    elif is_namedtuple(value):
        return deform_to_bus(dict(value._asdict()))
    elif is_dataclass(value):
        from dataclasses import asdict

        return deform_to_bus(asdict(value))
    elif isinstance_safe(value, frozendict):
        return deform_to_bus(value._dict)
    elif isinstance_safe(value, Enum):
        return deform_to_bus(value.value)
    elif isinstance_safe(value, (datetime, date)):
        return value.isoformat()
    elif isinstance_safe(value, UUID):
        return str(value)
    elif isinstance_safe(value, set):
        return [deform_to_bus(v) for v in value]
    elif type(value) == tuple:
        return [deform_to_bus(v) for v in value]
    elif isinstance_safe(value, list):
        return [deform_to_bus(v) for v in value]
    elif isinstance_safe(value, (int, float, str)):
        return value
    elif isinstance_safe(value, (bytes, memoryview)):
        return b64encode(bytes(value)).decode("utf8")
    elif isinstance_safe(value, (Decimal, complex)):
        return str(value)
    elif hasattr(value, "__module__"):
        # some kind of custom object we don't recognise
        raise DeformError(
            f"Failed to deform value of type {type(value)} for "
            f"transmission on the bus. Perhaps specify the "
            f"__to_bus__() and __from_bus__() methods on the class. Alternatively, "
            f"transform the data before placing it onto the bus."
        )
    else:
        # A built-in that we missed in the above checks?
        return value



from typing import Optional, TYPE_CHECKING, Any, Generator

from lightbus.client.utilities import OnError
from lightbus.exceptions import InvalidBusPathConfiguration, InvalidParameters
from lightbus.utilities.async_tools import block

if TYPE_CHECKING:
    # pylint: disable=unused-import,cyclic-import
    from lightbus import BusClient

__all__ = ["BusPath"]


class BusPath:
    """Represents a path on the bus

    This class provides a higher-level wrapper around the `BusClient` class.
    This wrapper allows for a more idiomatic use of the bus. For example:

        bus.auth.get_user(username='admin')

    Compare this to the lower level equivalent using the `BusClient`:

        bus.client.call_rpc_remote(
            api_name='auth',
            name='get_user',
            kwargs={'username': 'admin'},
        )

    """

    def __init__(self, name: str, *, parent: Optional["BusPath"], client: "BusClient"):
        if not parent and name:
            raise InvalidBusPathConfiguration("Root client node may not have a name")
        self.name = name
        self.parent = parent
        self.client = client

    def __getattr__(self, item) -> "BusPath":
        return self.__class__(name=item, parent=self, client=self.client)

    def __str__(self):
        return self.fully_qualified_name

    def __repr__(self):
        return "<BusPath {}>".format(self.fully_qualified_name)

    def __dir__(self):
        # Used by `lightbus shell` command
        path = [node.name for node in self.ancestors(include_self=True)]
        path.reverse()

        api_names = [[""] + n.split(".") for n in self.client.api_registry.names()]

        matches = []
        apis = []
        for api_name in api_names:
            if api_name == path:
                # Api name matches exactly
                apis.append(api_name)
            elif api_name[: len(path)] == path:
                # Partial API match
                matches.append(api_name[len(path)])

        for api_name in apis:
            api = self.client.api_registry.get(".".join(api_name[1:]))
            matches.extend(dir(api))

        return matches

    # RPC

    def __call__(self, *args, **kwargs):
        """Call this BusPath node as an RPC"""
        return self.call(*args, **kwargs)

    def call(self, *args, bus_options: dict = None, **kwargs):
        """Call this BusPath node as an RPC"

        In contrast to __call__(), this method provides the ability to call
        with the additional `bus_options` argument.
        """
        # Use a larger value of `rpc_timeout` because call_rpc_remote() should
        # handle timeout
        rpc_timeout = self.client.config.api(self.api_name).rpc_timeout * 1.5
        return block(self.call_async(*args, **kwargs, bus_options=bus_options), timeout=rpc_timeout)

    async def call_async(self, *args, bus_options=None, **kwargs):
        """Call this BusPath node as an RPC (asynchronous)"

        In contrast to __call__(), this method provides the ability to call
        with the additional `bus_options` argument.
        """
        if args:
            raise InvalidParameters(
                f"You have attempted to call the RPC {self.fully_qualified_name} using positional "
                f"arguments. Lightbus requires you use keyword arguments. For example, "
                f"instead of func(1), use func(foo=1)."
            )

        bus_options = bus_options or {}
        return await self.client.call_rpc_remote(
            api_name=self.api_name, name=self.name, kwargs=kwargs, options=bus_options
        )

    # Events

    def listen(
        self,
        listener,
        *,
        listener_name: str,
        bus_options: dict = None,
        on_error: OnError = OnError.SHUTDOWN,
    ):
        """Listen to events for this BusPath node"""
        return self.client.listen_for_event(
            api_name=self.api_name,
            name=self.name,
            listener=listener,
            listener_name=listener_name,
            options=bus_options,
            on_error=on_error,
        )

    def fire(self, *args, bus_options: dict = None, **kwargs):
        """Fire an event for this BusPath node"""
        return block(
            self.fire_async(*args, **kwargs, bus_options=bus_options),
            timeout=self.client.config.api(self.api_name).event_fire_timeout,
        )

    async def fire_async(self, *args, bus_options: dict = None, **kwargs):
        """Fire an event for this BusPath node (asynchronous)"""
        if args:
            raise InvalidParameters(
                f"You have attempted to fire the event {self.fully_qualified_name} using positional "
                f"arguments. Lightbus requires you use keyword arguments. For example, "
                f"instead of func(1), use func(foo=1)."
            )
        return await self.client.fire_event(
            api_name=self.api_name, name=self.name, kwargs=kwargs, options=bus_options
        )

    # Utilities

    def ancestors(self, include_self=False) -> Generator["BusPath", None, None]:
        """Get all ancestors of this node"""
        parent = self
        while parent is not None:
            if parent != self or include_self:
                yield parent
            parent = parent.parent

    @property
    def api_name(self) -> str:
        """Get the API name of this node

        This assumes the full path to this node is a fully qualified event/rpc name
        """
        path = [node.name for node in self.ancestors(include_self=False)]
        path.reverse()
        return ".".join(path[1:])

    @property
    def fully_qualified_name(self) -> str:
        """Get the fully qualified string name of this node
        """
        path = [node.name for node in self.ancestors(include_self=True)]
        path.reverse()
        return ".".join(path[1:])

    # Schema

    @property
    def schema(self):
        """Get the bus schema"""
        if self.parent is None:
            return self.client.schema
        else:
            # TODO: Implement getting schema of child nodes if there is demand
            raise AttributeError(
                "Schema only available on root node. Use bus.schema, not bus.my_api.schema"
            )

    @property
    def parameter_schema(self):
        """Get the parameter JSON schema for the given event or RPC"""
        # TODO: Test
        return self.client.schema.get_event_or_rpc_schema(self.api_name, self.name)["parameters"]

    @property
    def response_schema(self):
        """Get the response JSON schema for the given RPC

        Only RPCs have responses. Accessing this property for an event will result in a
        SchemaNotFound error.
        """
        return self.client.schema.get_rpc_schema(self.api_name, self.name)["response"]

    def validate_parameters(self, parameters: dict):
        """Validate the parameters for an event or RPC against the schema

        See Also: https://lightbus.org/reference/schema/
        """
        self.client.schema.validate_parameters(self.api_name, self.name, parameters)

    def validate_response(self, response: Any):
        """Validate the response for an RPC against the schema

        See Also: https://lightbus.org/reference/schema/
        """
        self.client.schema.validate_parameters(self.api_name, self.name, response)



import argparse
import logging

import sys
from pathlib import Path

from lightbus.commands import utilities as command_utilities
from lightbus.plugins import PluginRegistry
from lightbus.utilities.async_tools import block

logger = logging.getLogger(__name__)


class Command:
    def setup(self, parser, subparsers):
        parser_dumpschema = subparsers.add_parser(
            "dumpschema",
            help="Dumps all currently present bus schemas to a file or directory",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        group = parser_dumpschema.add_argument_group(title="Dump config schema command arguments")
        group.add_argument(
            "--out",
            "-o",
            help=(
                "File or directory to write schema to. If a directory is "
                "specified one schema file will be created for each API. "
                "If omitted the schema will be written to standard out."
            ),
            metavar="FILE_OR_DIRECTORY",
        )
        command_utilities.setup_common_arguments(parser_dumpschema)
        parser_dumpschema.set_defaults(func=self.handle)

    def handle(self, args, config, plugin_registry: PluginRegistry):
        command_utilities.setup_logging(args.log_level or "warning", config)

        bus_module, bus = command_utilities.import_bus(args)
        block(bus.client.lazy_load_now())
        bus.schema.save_local(args.out)

        if args.out:
            sys.stderr.write(
                "Schema for {} APIs saved to {}\n".format(
                    len(bus.schema.api_names), Path(args.out).resolve()
                )
            )



""" Configuration structure generation

We do some pretty exciting things here to generate the configuration
structure.

We load in all available plugins and transports and get the desired
configuration format for each (via the parameters of `from_config()` in both
cases). From there we dynamically generate named tuples which form
the structure for the configuration.

Benefits of this are:

    1. Many configuration errors can be caught early
    2. We can use the schema-related code to generate a JSON schema for the config.
       This can be used for validation, auto completion, and improving tooling in general
       (smart config editing in the management UI?)
    3. Plugins and transports can be assured of reasonable sane data

See Also: https://lightbus.org/reference/configuration/
"""
import logging
import os
import socket
from enum import Enum
from typing import NamedTuple, Optional, Union, Dict

from lightbus.plugins import find_plugins
from lightbus.transports.registry import get_available_transports
from lightbus.utilities.config import random_name
from lightbus.utilities.human import generate_human_friendly_name

logger = logging.getLogger(__name__)

# pylint: disable=exec-used


def make_transport_selector_structure(type_) -> NamedTuple:
    class_name = f"{type_.title()}TransportSelector"
    code = f"class {class_name}(NamedTuple):\n    pass\n"
    variables = {}
    transports = get_available_transports(type_)
    for transport_name, transport_class in transports.items():
        variables[transport_class.__name__] = transport_class
        code += f"    {transport_name}: Optional[{transport_class.__name__}.Config] = None\n"

    globals_ = globals().copy()
    globals_.update(variables)
    exec(code, globals_)  # nosec
    return globals_[class_name]


def make_plugin_selector_structure() -> NamedTuple:
    class_name = f"PluginSelector"
    code = f"class {class_name}(NamedTuple):\n    pass\n"
    variables = {}

    for plugin_name, plugin_class in find_plugins().items():
        plugin_class_name = plugin_class.__name__
        variables[plugin_class_name] = plugin_class
        code += f"    {plugin_name}: Optional[{plugin_class_name}.Config] = {plugin_class_name}.Config()\n"

    globals_ = globals().copy()
    globals_.update(variables)
    exec(code, globals_)  # nosec
    return globals_[class_name]


RpcTransportSelector = make_transport_selector_structure("rpc")
ResultTransportSelector = make_transport_selector_structure("result")
EventTransportSelector = make_transport_selector_structure("event")
SchemaTransportSelector = make_transport_selector_structure("schema")

PluginSelector = make_plugin_selector_structure()


class LogLevelEnum(Enum):
    DEBUG = "debug"
    INFO = "info"
    WARNING = "warning"
    ERROR = "error"
    CRITICAL = "critical"


class ApiValidationConfig(NamedTuple):
    outgoing: bool = True
    incoming: bool = True


class ApiConfig:
    rpc_timeout: int = 5
    event_listener_setup_timeout: int = 1
    event_fire_timeout: int = 5
    validate: Optional[Union[ApiValidationConfig, bool]] = ApiValidationConfig()
    event_transport: EventTransportSelector = None
    rpc_transport: RpcTransportSelector = None
    result_transport: ResultTransportSelector = None
    strict_validation: bool = False
    #: Cast values before calling event listeners and RPCs
    cast_values: bool = True

    def __init__(self, **kw):
        for k, v in kw.items():
            setattr(self, k, v)

        self._normalise_validate()

    @classmethod
    def __from_bus__(cls, data: dict):
        data = {k: v for k, v in data.items() if k in dir(cls)}
        return cls(**data)

    def _normalise_validate(self):
        if self.validate in (True, False):
            # Expand out the true/false shortcut
            self.validate = ApiValidationConfig(outgoing=self.validate, incoming=self.validate)
        elif isinstance(self.validate, dict):
            # pylint: disable=not-a-mapping
            self.validate = ApiValidationConfig(**self.validate)


class SchemaConfig(NamedTuple):
    human_readable: bool = True
    ttl: int = 60
    transport: SchemaTransportSelector = None


class BusConfig(NamedTuple):
    log_level: LogLevelEnum = LogLevelEnum.INFO
    schema: SchemaConfig = SchemaConfig()


class RootConfig:
    service_name: str = "{friendly}"
    process_name: str = "{random4}"
    bus: BusConfig = BusConfig()
    apis: Dict[str, ApiConfig] = {}
    plugins: PluginSelector = PluginSelector()

    def __init__(self, **kw):
        for k, v in kw.items():
            setattr(self, k, v)

        self.set_service_name(self.service_name)
        self.set_process_name(self.process_name)

    @classmethod
    def __from_bus__(cls, data: dict):
        data = {k: v for k, v in data.items() if k in dir(cls)}
        return cls(**data)

    def set_service_name(self, name):
        self.service_name = self._format_name(name)

    def set_process_name(self, name):
        self.process_name = self._format_name(name)

    def _format_name(self, name):
        random_string = random_name(length=16)
        return name.format(
            hostname=socket.gethostname().lower(),
            pid=os.getpid(),
            random4=random_string[:4],
            random8=random_string[:8],
            random16=random_string[:16],
            friendly=generate_human_friendly_name(),
        )



import asyncio

from lightbus.client.internal_messaging.consumer import InternalConsumer
from lightbus.client.internal_messaging.producer import InternalProducer
from lightbus.api import ApiRegistry
from lightbus.client.utilities import ErrorQueueType
from lightbus.config import Config
from lightbus.transports.registry import TransportRegistry
from lightbus.utilities.internal_queue import InternalQueue


class BaseDock:
    """The base dock

    A dock is responsible for interfacing a transport with Lightbus' internal
    messaging system.
    """

    def __init__(
        self,
        transport_registry: TransportRegistry,
        api_registry: ApiRegistry,
        config: Config,
        error_queue: ErrorQueueType,
        consume_from: InternalQueue,
        produce_to: InternalQueue,
    ):
        self.transport_registry = transport_registry
        self.api_registry = api_registry
        self.config = config
        self.error_queue = error_queue
        self.producer = InternalProducer(queue=produce_to, error_queue=error_queue)
        self.consumer = InternalConsumer(queue=consume_from, error_queue=error_queue)

        self.producer.start()
        self.consumer.start(self.handle)

    async def handle(self, command):
        raise NotImplementedError()

    async def wait_until_ready(self):
        await self.producer.wait_until_ready()
        await self.consumer.wait_until_ready()



"""Utility functions relating to bus creation"""
import logging
import os
import sys
import threading
import weakref
from functools import partial
from typing import Union, Mapping, Type, List, Optional, Callable

from lightbus import Schema
from lightbus.api import ApiRegistry
from lightbus.client.docks.event import EventDock
from lightbus.client.docks.rpc_result import RpcResultDock
from lightbus.client.subclients.event import EventClient
from lightbus.client.subclients.rpc_result import RpcResultClient
from lightbus.client.utilities import ErrorQueueType
from lightbus.hooks import HookRegistry
from lightbus.internal_apis import LightbusStateApi, LightbusMetricsApi
from lightbus.plugins import PluginRegistry
from lightbus.utilities.async_tools import get_event_loop, block
from lightbus.utilities.internal_queue import InternalQueue
from lightbus.utilities.features import ALL_FEATURES, Feature
from lightbus.config.structure import RootConfig
from lightbus.log import Bold
from lightbus.path import BusPath
from lightbus.client.bus_client import BusClient
from lightbus.config import Config
from lightbus.exceptions import FailedToImportBusModule
from lightbus.transports.registry import TransportRegistry
from lightbus.utilities.importing import import_module_from_string
from lightbus.utilities.logging import log_welcome_message

__all__ = ["create", "load_config", "import_bus_module", "get_bus"]
logger = logging.getLogger(__name__)


def create_client(
    config: Union[dict, RootConfig] = None,
    *,
    config_file: str = None,
    service_name: str = None,
    process_name: str = None,
    features: List[Union[Feature, str]] = ALL_FEATURES,
    client_class: Type[BusClient] = BusClient,
    node_class: Type[BusPath] = BusPath,
    plugins=None,
    flask: bool = False,
    hook_registry: Optional[HookRegistry] = None,
    api_registry: Optional[ApiRegistry] = None,
    **kwargs,
) -> BusClient:
    """
    Create a new bus instance which can be used to access the bus.

    Typically this will be used as follows:

        import lightbus

        bus = lightbus.create()

    This will be a `BusPath` instance. If you wish to access the lower
    level `BusClient` you can do so via `bus.client`.

    Args:
        config (dict, Config): The config object or dictionary to load
        config_file (str): The path to a config file to load (should end in .json or .yaml)
        service_name (str): The name of this service - will be used when creating event consumer groups
        process_name (str): The unique name of this process - used when retrieving unprocessed events following a crash
        client_class (Type[BusClient]): The class from which the bus client will be instantiated
        node_class (BusPath): The class from which the bus path will be instantiated
        plugins (list): A list of plugin instances to load
        flask (bool): Are we using flask? If so we will make sure we don't start lightbus in the reloader process
        **kwargs (): Any additional instantiation arguments to be passed to `client_class`.

    Returns: BusPath

    """
    if flask:
        in_flask_server = sys.argv[0].endswith("flask") and "run" in sys.argv
        if in_flask_server and os.environ.get("WERKZEUG_RUN_MAIN", "").lower() != "true":
            # Flask has a reloader process that shouldn't start a lightbus client
            return

    # Ensure an event loop exists, as creating InternalQueue
    # objects requires that we have one.
    get_event_loop()

    # If were are running via the Lightbus CLI then we may have
    # some command line arguments we need to apply.
    # pylint: disable=cyclic-import,import-outside-toplevel
    from lightbus.commands import COMMAND_PARSED_ARGS

    config_file = COMMAND_PARSED_ARGS.get("config_file", None) or config_file
    service_name = COMMAND_PARSED_ARGS.get("service_name", None) or service_name
    process_name = COMMAND_PARSED_ARGS.get("process_name", None) or process_name

    if config is None:
        config = load_config(
            from_file=config_file, service_name=service_name, process_name=process_name
        )

    if isinstance(config, Mapping):
        config = Config.load_dict(config or {})
    elif isinstance(config, RootConfig):
        config = Config(config)

    transport_registry = kwargs.pop("transport_registry", None) or TransportRegistry().load_config(
        config
    )

    schema = Schema(
        schema_transport=transport_registry.get_schema_transport(),
        max_age_seconds=config.bus().schema.ttl,
        human_readable=config.bus().schema.human_readable,
    )

    error_queue: ErrorQueueType = InternalQueue()

    # Plugin registry

    plugin_registry = PluginRegistry()
    if plugins is None:
        logger.debug("Auto-loading any installed Lightbus plugins...")
        plugin_registry.autoload_plugins(config)
    else:
        logger.debug("Loading explicitly specified Lightbus plugins....")
        plugin_registry.set_plugins(plugins)

    # Hook registry
    if not hook_registry:
        hook_registry = HookRegistry(
            error_queue=error_queue, execute_plugin_hooks=plugin_registry.execute_hook
        )

    # API registry
    if not api_registry:
        api_registry = ApiRegistry()
        api_registry.add(LightbusStateApi())
        api_registry.add(LightbusMetricsApi())

    events_queue_client_to_dock = InternalQueue()
    events_queue_dock_to_client = InternalQueue()

    event_client = EventClient(
        api_registry=api_registry,
        hook_registry=hook_registry,
        config=config,
        schema=schema,
        error_queue=error_queue,
        consume_from=events_queue_dock_to_client,
        produce_to=events_queue_client_to_dock,
    )

    event_dock = EventDock(
        transport_registry=transport_registry,
        api_registry=api_registry,
        config=config,
        error_queue=error_queue,
        consume_from=events_queue_client_to_dock,
        produce_to=events_queue_dock_to_client,
    )

    rpcs_queue_client_to_dock = InternalQueue()
    rpcs_queue_dock_to_client = InternalQueue()

    rpc_result_client = RpcResultClient(
        api_registry=api_registry,
        hook_registry=hook_registry,
        config=config,
        schema=schema,
        error_queue=error_queue,
        consume_from=rpcs_queue_dock_to_client,
        produce_to=rpcs_queue_client_to_dock,
    )

    rpc_result_dock = RpcResultDock(
        transport_registry=transport_registry,
        api_registry=api_registry,
        config=config,
        error_queue=error_queue,
        consume_from=rpcs_queue_client_to_dock,
        produce_to=rpcs_queue_dock_to_client,
    )

    client = client_class(
        config=config,
        hook_registry=hook_registry,
        plugin_registry=plugin_registry,
        features=features,
        schema=schema,
        api_registry=api_registry,
        event_client=event_client,
        rpc_result_client=rpc_result_client,
        error_queue=error_queue,
        transport_registry=transport_registry,
        **kwargs,
    )

    # Pass the client to any hooks
    # (use a weakref to prevent circular references)
    hook_registry.set_extra_parameter("client", weakref.proxy(client))

    # We don't do this normally as the docks do not need to be
    # accessed directly, but this is useful in testing
    # TODO: Testing flag removed, but these are only needed in testing.
    #       Perhaps wrap them up in a way that makes this obvious
    client.event_dock = event_dock
    client.rpc_result_dock = rpc_result_dock

    log_welcome_message(
        logger=logger,
        transport_registry=transport_registry,
        schema=schema,
        plugin_registry=plugin_registry,
        config=config,
    )

    # Wait until the docks are ready to go
    if not get_event_loop().is_running():
        # If the event loop is running then we can assume that this setup
        # will happen momentarily anyway. Plus, if an event loop is running
        # then we cannot block. This is mainly to make sure everything
        # starts up in a sensible fashion when we are in non-worker-mode.
        block(event_dock.wait_until_ready(), timeout=2)
        block(rpc_result_dock.wait_until_ready(), timeout=2)

    return client


def create(
    config: Union[dict, RootConfig] = None,
    *,
    config_file: str = None,
    service_name: str = None,
    process_name: str = None,
    features: List[Union[Feature, str]] = ALL_FEATURES,
    client_class: Type[BusClient] = BusClient,
    node_class: Type[BusPath] = BusPath,
    plugins=None,
    flask: bool = False,
    **kwargs,
) -> BusPath:
    client_proxy = ThreadLocalClientProxy(
        partial(
            create_client,
            config,
            config_file=config_file,
            service_name=service_name,
            process_name=process_name,
            features=features,
            client_class=client_class,
            node_class=node_class,
            plugins=plugins,
            flask=flask,
            **kwargs,
        )
    )

    return node_class(name="", parent=None, client=client_proxy)


class ThreadLocalClientProxy:
    """A proxy which wraps around the BusClient

    This proxy will ensure a bus client is created for each thread.
    """

    def __init__(self, client_factory: Callable):
        self.client_factory = client_factory
        self.local = threading.local()
        # Create the client from the main thread. We'll assume the current
        # thread is the main thread, even if that is not strictly the case
        self.main_client: Optional[BusClient] = self.client_factory()
        # Make the main client available in the current thread's
        # local storage. We do this because the current thread is the 'main' thread
        self.local.client = self.main_client
        self.enabled = True

    def _create_client(self) -> BusClient:
        """Create a bus client using our client_factory"""
        # Create another client, as the bus has been accessed
        # from a thread. Pass this new client the hook & api
        # registry which would have been configured post-instantiation,
        # and therefore our client_factory will not be aware of this state.
        # TODO: Consider copying this rather than passing a reference
        return self.client_factory(
            hook_registry=self.main_client.hook_registry,
            api_registry=self.main_client.api_registry,
        )

    @property
    def proxied_client(self):
        """Directly access the BusClient instance

        Useful in some edge-cases, such as mocking within the tests
        """
        if not self.enabled:
            return self.main_client

        if not hasattr(self.local, "client"):
            logger.debug(f"Creating new client for thread {threading.current_thread().name}")
            self.local.client = self._create_client()
        return self.local.client

    def disable_proxy(self):
        """Disable the proxying and always use self.main_client

        This is mainly useful within the lightbus worker (i.e. `lightbus run`).
        In this case lightbus is in control of the process so can make sane choices
        with regards its use of threads, rather than being at the mercy of some
        other system (i.e. gunicorn, django's development server, etc)
        """
        self.enabled = False

    def __getattr__(self, item):
        return getattr(self.proxied_client, item)


def load_config(
    from_file: str = None, service_name: str = None, process_name: str = None, quiet: bool = False
) -> Config:
    from_file = from_file or os.environ.get("LIGHTBUS_CONFIG")

    if from_file:
        if not quiet:
            logger.info(f"Loading config from {from_file}")
        config = Config.load_file(file_path=from_file)
    else:
        if not quiet:
            logger.info(f"No config file specified, will use default config")
        config = Config.load_dict({})

    if service_name:
        config._config.set_service_name(service_name)

    if process_name:
        config._config.set_process_name(process_name)

    return config


def import_bus_module(bus_module_name: str = None):
    """Get the bus module, importing if necessary

    The bus module is determined as follows:

      1. The bus_module_name, if present
      2. The `LIGHTBUS_MODULE` environment variable, if present
      3. Otherwise, the default of `bus` will be used
    """
    bus_module_name = bus_module_name or os.environ.get("LIGHTBUS_MODULE", "") or "bus"
    logger.info(f"Importing bus.py from {Bold(bus_module_name)}")
    try:
        bus_module = import_module_from_string(bus_module_name)
    except ImportError as e:
        raise FailedToImportBusModule(
            f"Failed to import bus module '{bus_module_name}'.\n\n"
            "    Perhaps you need to set the LIGHTBUS_MODULE environment variable? Alternatively "
            "you may need to configure your PYTHONPATH.\n\n"
            f"    Error was: {e}"
        )

    if not hasattr(bus_module, "bus"):
        raise FailedToImportBusModule(
            f"Bus module at '{bus_module_name}' contains no attribute named 'bus'.\n\n"
            "    Your bus module must set a variable named bus using:\n"
            "        bus = lightbus.create()"
        )

    if not isinstance(bus_module.bus, BusPath):
        raise FailedToImportBusModule(
            f"Bus module at '{bus_module_name}' contains an invalid value for the 'bus'"
            " attribute.\n\n    We expected a BusPath instance, but found"
            f" '{type(bus_module.bus).__name__}'.\n    Your bus module must contain a variable"
            " named 'bus' using:\n        bus = lightbus.create()"
        )

    return bus_module


def get_bus(bus_module_name: str = None) -> BusPath:
    """Get the bus, importing if necessary"""
    return import_bus_module(bus_module_name).bus



import logging

from typing import NamedTuple, Optional, List, Tuple

from lightbus.api import Api
from lightbus.message import EventMessage, RpcMessage, ResultMessage
from lightbus.utilities.internal_queue import InternalQueue

logger = logging.getLogger(__name__)


class SendEventCommand(NamedTuple):
    message: EventMessage
    options: dict = {}


class ConsumeEventsCommand(NamedTuple):
    events: List[Tuple[str, str]]
    listener_name: str
    destination_queue: InternalQueue[EventMessage]
    options: dict = {}


class AcknowledgeEventCommand(NamedTuple):
    message: EventMessage
    options: dict = {}


class CallRpcCommand(NamedTuple):
    message: RpcMessage
    options: dict = {}


class ConsumeRpcsCommand(NamedTuple):
    api_names: List[str]
    options: dict = {}


class ExecuteRpcCommand(NamedTuple):
    """An RPC call has been received and must be executed locally"""

    message: RpcMessage


class PublishApiSchemaCommand(NamedTuple):
    api: Api


class CloseCommand(NamedTuple):
    pass


class SendResultCommand(NamedTuple):
    rpc_message: RpcMessage
    message: ResultMessage


class ReceiveResultCommand(NamedTuple):
    message: RpcMessage
    destination_queue: InternalQueue
    options: dict


class ReceiveSchemaUpdateCommand(NamedTuple):
    schema: dict


class ShutdownCommand(NamedTuple):
    exception: Optional[BaseException]



"""
An XPath for JSON

A port of the Perl, and JavaScript versions of JSONPath
see http://goessner.net/articles/JsonPath/

Based on on JavaScript version by Stefan Goessner at:
        https://goessner.net/articles/JsonPath/
        http://code.google.com/p/jsonpath/
and Perl version by Kate Rhodes at:
        http://github.com/masukomi/jsonpath-perl/tree/master

Python3 compatibily by Per J. Sandstrom
"""
from __future__ import print_function

__author__ = "Phil Budne"
__revision__ = "$Revision: 1.16 $"
__version__ = "0.81"

# 	Copyright (c) 2007 Stefan Goessner (goessner.net)
#       Copyright (c) 2008 Kate Rhodes (masukomi.org)
#       Copyright (c) 2008-2018 Philip Budne (ultimate.com)
# 	Licensed under the MIT licence:
#
# 	Permission is hereby granted, free of charge, to any person
# 	obtaining a copy of this software and associated documentation
# 	files (the "Software"), to deal in the Software without
# 	restriction, including without limitation the rights to use,
# 	copy, modify, merge, publish, distribute, sublicense, and/or sell
# 	copies of the Software, and to permit persons to whom the
# 	Software is furnished to do so, subject to the following
# 	conditions:
#
# 	The above copyright notice and this permission notice shall be
# 	included in all copies or substantial portions of the Software.
#
# 	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# 	EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# 	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# 	NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# 	HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# 	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# 	FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# 	OTHER DEALINGS IN THE SOFTWARE.

import re
import sys

# XXX BUGS:
# evalx is generally a crock:
#       handle !@.name.name???
# there are probably myriad unexpected ways to get an exception:
#       wrap initial "trace" call in jsonpath body in a try/except??

# XXX TODO:
# internally keep paths as lists to preserve integer types
#       (instead of as ';' delimited strings)

__all__ = ["jsonpath"]

# XXX precompile RE objects on load???
# re_1 = re.compile(.....)
# re_2 = re.compile(.....)

# For python3 portability
if sys.version_info[0] == 3:
    xrange = range


def normalize(x):
    """normalize the path expression; outside jsonpath to allow testing"""
    subx = []

    # replace index/filter expressions with placeholders
    # Python anonymous functions (lambdas) are cryptic, hard to debug
    def f1(m):
        n = len(subx)  # before append
        g1 = m.group(1)
        subx.append(g1)
        ret = "[#%d]" % n
        #       print("f1:", g1, ret)
        return ret

    x = re.sub(r"[\['](\??\(.*?\))[\]']", f1, x)

    # added the negative lookbehind -krhodes
    x = re.sub(r"'?(?<!@)\.'?|\['?", ";", x)

    x = re.sub(r";;;|;;", ";..;", x)

    x = re.sub(r";$|'?\]|'$", "", x)

    # put expressions back
    def f2(m):
        g1 = m.group(1)
        #       print("f2:", g1)
        return subx[int(g1)]

    x = re.sub(r"#([0-9]+)", f2, x)

    return x


def jsonpath(obj, expr, result_type="VALUE", debug=0, use_eval=True):
    """traverse JSON object using jsonpath expr, returning values or paths"""

    def s(x, y):
        """concatenate path elements"""
        return str(x) + ";" + str(y)

    def isint(x):
        """check if argument represents a decimal integer"""
        return x.isdigit()

    def as_path(path):
        """convert internal path representation to
           "full bracket notation" for PATH output"""
        p = "$"
        for piece in path.split(";")[1:]:
            # make a guess on how to index
            # XXX need to apply \ quoting on '!!
            if isint(piece):
                p += "[%s]" % piece
            else:
                p += "['%s']" % piece
        return p

    def store(path, object):
        if result_type == "VALUE":
            result.append(object)
        elif result_type == "IPATH":  # Index format path (Python ext)
            # return list of list of indices -- can be used w/o "eval" or split
            result.append(path.split(";")[1:])
        else:  # PATH
            result.append(as_path(path))
        return path

    def trace(expr, obj, path):
        if debug:
            print("trace", expr, "/", path)
        if expr:
            x = expr.split(";")
            loc = x[0]
            x = ";".join(x[1:])
            if debug:
                print("\t", loc, type(obj))
            if loc == "*":

                def f03(key, loc, expr, obj, path):
                    if debug > 1:
                        print("\tf03", key, loc, expr, path)
                    trace(s(key, expr), obj, path)

                walk(loc, x, obj, path, f03)
            elif loc == "..":
                trace(x, obj, path)

                def f04(key, loc, expr, obj, path):
                    if debug > 1:
                        print("\tf04", key, loc, expr, path)
                    if isinstance(obj, dict):
                        if key in obj:
                            trace(s("..", expr), obj[key], s(path, key))
                    else:
                        if key < len(obj):
                            trace(s("..", expr), obj[key], s(path, key))

                walk(loc, x, obj, path, f04)
            elif loc == "!":
                # Perl jsonpath extension: return keys
                def f06(key, loc, expr, obj, path):
                    if isinstance(obj, dict):
                        trace(expr, key, path)

                walk(loc, x, obj, path, f06)
            elif isinstance(obj, dict) and loc in obj:
                trace(x, obj[loc], s(path, loc))
            elif isinstance(obj, list) and isint(loc):
                iloc = int(loc)
                if debug:
                    print("----->", iloc, len(obj))
                if len(obj) > iloc:
                    trace(x, obj[iloc], s(path, loc))
            else:
                # [(index_expression)]
                if loc.startswith("(") and loc.endswith(")"):
                    if debug > 1:
                        print("index", loc)
                    e = evalx(loc, obj)
                    trace(s(e, x), obj, path)
                    return

                # ?(filter_expression)
                if loc.startswith("?(") and loc.endswith(")"):
                    if debug > 1:
                        print("filter", loc)

                    def f05(key, loc, expr, obj, path):
                        if debug > 1:
                            print("f05", key, loc, expr, path)
                        if isinstance(obj, dict):
                            eval_result = evalx(loc, obj[key])
                        else:
                            eval_result = evalx(loc, obj[int(key)])
                        if eval_result:
                            trace(s(key, expr), obj, path)

                    loc = loc[2:-1]
                    walk(loc, x, obj, path, f05)
                    return

                m = re.match(r"(-?[0-9]*):(-?[0-9]*):?(-?[0-9]*)$", loc)
                if m:
                    if isinstance(obj, (dict, list)):

                        def max(x, y):
                            if x > y:
                                return x
                            return y

                        def min(x, y):
                            if x < y:
                                return x
                            return y

                        objlen = len(obj)
                        s0 = m.group(1)
                        s1 = m.group(2)
                        s2 = m.group(3)

                        # XXX int("badstr") raises exception
                        start = int(s0) if s0 else 0
                        end = int(s1) if s1 else objlen
                        step = int(s2) if s2 else 1

                        if start < 0:
                            start = max(0, start + objlen)
                        else:
                            start = min(objlen, start)
                        if end < 0:
                            end = max(0, end + objlen)
                        else:
                            end = min(objlen, end)

                        for i in xrange(start, end, step):
                            trace(s(i, x), obj, path)
                    return

                # after (expr) & ?(expr)
                if loc.find(",") >= 0:
                    # [index,index....]
                    for piece in re.split(r"'?,'?", loc):
                        if debug > 1:
                            print("piece", piece)
                        trace(s(piece, x), obj, path)
        else:
            store(path, obj)

    def walk(loc, expr, obj, path, funct):
        if isinstance(obj, list):
            for i in xrange(0, len(obj)):
                funct(i, loc, expr, obj, path)
        elif isinstance(obj, dict):
            for key in obj:
                funct(key, loc, expr, obj, path)

    def evalx(loc, obj):
        """eval expression"""

        if debug:
            print("evalx", loc)

        # a nod to JavaScript. doesn't work for @.name.name.length
        # Write len(@.name.name) instead!!!
        loc = loc.replace("@.length", "len(__obj)")

        loc = loc.replace("&&", " and ").replace("||", " or ")

        # replace !@.name with 'name' not in obj
        # XXX handle !@.name.name.name....
        def notvar(m):
            return "'%s' not in __obj" % m.group(1)

        loc = re.sub("!@\.([a-zA-Z@_]+)", notvar, loc)

        # replace @.name.... with __obj['name']....
        # handle @.name[.name...].length
        def varmatch(m):
            def brackets(elts):
                ret = "__obj"
                for e in elts:
                    if isint(e):
                        ret += "[%s]" % e  # ain't necessarily so
                    else:
                        ret += "['%s']" % e  # XXX beware quotes!!!!
                return ret

            g1 = m.group(1)
            elts = g1.split(".")
            if elts[-1] == "length":
                return "len(%s)" % brackets(elts[1:-1])
            return brackets(elts[1:])

        loc = re.sub(r"(?<!\\)(@\.[a-zA-Z@_.]+)", varmatch, loc)

        # removed = -> == translation
        # causes problems if a string contains =

        # replace @  w/ "__obj", but \@ means a literal @
        loc = re.sub(r"(?<!\\)@", "__obj", loc).replace(r"\@", "@")
        if not use_eval:
            if debug:
                print("eval disabled")
            raise Exception("eval disabled")
        if debug:
            print("eval", loc)
        try:
            # eval w/ caller globals, w/ local "__obj"!
            v = eval(loc, caller_globals, {"__obj": obj})
        except Exception as e:
            if debug:
                print(repr(e))
            return False

        if debug:
            print("->", v)
        return v

    # body of jsonpath()

    # Get caller globals so eval can pick up user functions!!!
    caller_globals = sys._getframe(1).f_globals
    result = []
    if expr and obj:
        cleaned_expr = normalize(expr)
        if cleaned_expr.startswith("$;"):
            cleaned_expr = cleaned_expr[2:]

        # XXX wrap this in a try??
        trace(cleaned_expr, obj, "$")

        if len(result) > 0:
            return result
    return False


if __name__ == "__main__":
    try:
        import json  # v2.6
    except ImportError:
        import simplejson as json

    import sys

    # XXX take options for output format, output file, debug level

    if len(sys.argv) < 3 or len(sys.argv) > 4:
        sys.stdout.write("Usage: jsonpath.py FILE PATH [OUTPUT_TYPE]\n")
        sys.exit(1)

    object = json.load(file(sys.argv[1]))
    path = sys.argv[2]
    format = "VALUE"

    if len(sys.argv) > 3:
        # XXX verify?
        format = sys.argv[3]

    value = jsonpath(object, path, format)

    if not value:
        sys.exit(1)

    f = sys.stdout
    json.dump(value, f, sort_keys=True, indent=1)
    f.write("\n")

    sys.exit(0)



import asyncio
import threading
from functools import partial
from typing import TypeVar, Generic

T = TypeVar("T")


class InternalQueue(asyncio.Queue, Generic[T]):
    """Threadsafe version of asyncio.Queue"""

    def _init(self, maxsize):
        super()._init(maxsize)
        self._mutex = threading.RLock()
        # We will get the current loop dynamically
        # using asyncio.get_event_loop(), so make sure
        # we splat the parent classes loop to avoid using it
        self._loop = None

    def qsize(self):
        """Number of items in the queue."""
        with self._mutex:
            return len(self._queue)

    def empty(self):
        """Return True if the queue is empty, False otherwise."""
        with self._mutex:
            return not self._queue

    def full(self):
        """Return True if there are maxsize items in the queue.

        Note: if the Queue was initialized with maxsize=0 (the default),
        then full() is never True.
        """
        with self._mutex:
            if self._maxsize <= 0:
                return False
            else:
                return self.qsize() >= self._maxsize

    async def put(self, item: T):
        """Put an item into the queue.

        Put an item into the queue. If the queue is full, wait until a free
        slot is available before adding item.
        """

        with self._mutex:
            try:
                return self.put_nowait(item)
            except asyncio.QueueFull:
                # Queue is full, so start waiting until it has some
                # free space in the following while loop
                pass

            putter = asyncio.get_event_loop().create_future()
            self._putters.append(putter)

        while True:
            try:
                await putter
            except:
                with self._mutex:
                    putter.cancel()  # Just in case putter is not done yet.
                    try:
                        # Clean self._putters from canceled putters.
                        self._putters.remove(putter)
                    except ValueError:
                        # The putter could be removed from self._putters by a
                        # previous get_nowait call.
                        pass
                    if not self.full() and not putter.cancelled():
                        # We were woken up by get_nowait(), but can't take
                        # the call.  Wake up the next in line.
                        self._wakeup_next(self._putters)
                raise

            with self._mutex:
                try:
                    return self.put_nowait(item)
                except asyncio.QueueFull:
                    putter = asyncio.get_event_loop().create_future()
                    self._putters.append(putter)

    def put_nowait(self, item: T):
        """Put an item into the queue without blocking.

        If no free slot is immediately available, raise QueueFull.
        """
        with self._mutex:
            if self.full():
                raise asyncio.QueueFull
            self._put(item)
            self._unfinished_tasks += 1
            self._finished.clear()
            self._wakeup_next(self._getters)

    async def get(self) -> T:
        """Remove and return an item from the queue.

        If queue is empty, wait until an item is available.
        """
        with self._mutex:
            try:
                return self.get_nowait()
            except asyncio.QueueEmpty:
                pass

            getter = asyncio.get_event_loop().create_future()
            self._getters.append(getter)

        while True:
            try:
                await getter
            except:
                with self._mutex:
                    getter.cancel()  # Just in case getter is not done yet.
                    try:
                        # Clean self._getters from canceled getters.
                        self._getters.remove(getter)
                    except ValueError:
                        # The getter could be removed from self._getters by a
                        # previous put_nowait call.
                        pass
                    if not self.empty() and not getter.cancelled():
                        # We were woken up by put_nowait(), but can't take
                        # the call.  Wake up the next in line.
                        self._wakeup_next(self._getters)
                raise

            with self._mutex:
                try:
                    return self.get_nowait()
                except asyncio.QueueEmpty:
                    getter = asyncio.get_event_loop().create_future()
                    self._getters.append(getter)

    def get_nowait(self) -> T:
        """Remove and return an item from the queue.

        Return an item if one is immediately available, else raise QueueEmpty.
        """
        with self._mutex:
            if self.empty():
                raise asyncio.QueueEmpty
            item = self._get()
            self._wakeup_next(self._putters)
            return item

    def _wakeup_next(self, waiters):
        # Wake up the next waiter (if any) that isn't cancelled.
        while waiters:
            waiter = waiters.popleft()
            if not waiter.done():
                # We must set the result from within the waiter's
                # original event loop, so use call_soon_threadsafe()
                waiter.get_loop().call_soon_threadsafe(partial(waiter.set_result, None))
                break

    def task_done(self):
        """Indicate that a formerly enqueued task is complete.

        Used by queue consumers. For each get() used to fetch a task,
        a subsequent call to task_done() tells the queue that the processing
        on the task is complete.

        If a join() is currently blocking, it will resume when all items have
        been processed (meaning that a task_done() call was received for every
        item that had been put() into the queue).

        Raises ValueError if called more times than there were items placed in
        the queue.
        """
        with self._mutex:
            if self._unfinished_tasks <= 0:
                raise ValueError("task_done() called too many times")
            self._unfinished_tasks -= 1
            if self._unfinished_tasks == 0:
                self._finished.set()

    async def join(self):
        """Block until all items in the queue have been gotten and processed.

        The count of unfinished tasks goes up whenever an item is added to the
        queue. The count goes down whenever a consumer calls task_done() to
        indicate that the item was retrieved and all work on it is complete.
        When the count of unfinished tasks drops to zero, join() unblocks.
        """
        with self._mutex:
            if self._unfinished_tasks > 0:
                fut = self._finished.wait()
            else:
                fut = None

        if fut:
            await fut
